/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "plugin/plugin.pmh"

#define FE_REGISTRY_MANAGE_DEBUG	FALSE
#define FE_REGISTRY_MATCH_DEBUG		FALSE
#define FE_REGISTRY_FACTORY_DEBUG	FALSE
#define FE_REGISTRY_DUMP_LIBRARIES	(FE_CODEGEN<=FE_DEBUG)

namespace fe
{

Registry::Registry(void):
		m_name("Registry (no Master)"),
		m_tracker("Component")
{
	init();
}

Registry::Registry(Master* pMaster):
		m_name("Registry"),
		m_tracker("Component")
{
	if(pMaster)
		m_hpMaster=pMaster->getHandle();

	init();
}

void Registry::init(void)
{
#if FE_COUNTED_TRACK
	Counted::startTracker();
#endif
}

Registry::~Registry(void)
{
#if FE_REGISTRY_FACTORY_DEBUG
	feLog("~Registry\n");
	dump();
#endif

	m_dependents.clear();

#if FE_REGISTRY_DUMP_LIBRARIES
	feLog("~Registry libraries loaded:\n");
	dumpLibraries();
	feLog("\n");
#endif

	clear();
#if FE_COMPONENT_TRACK
	feLog("\nComponent %s\n",m_tracker.report().c_str());
#endif

	//* minimize Registry's impact on the report
	abandonHandle();
	m_hpMaster.invalidate();

#if FE_COUNTED_TRACK
	feLog("Counted %s\n",Counted::reportTracker().c_str());
	Counted::stopTracker();
#endif
}

Registry::LoadedLibrary::~LoadedLibrary(void)
{
	//* reuse chained library shutdown code (pComponent=NULL)
	fe::internal::checkUnload(m_spLibrary);
}

sp<Master> Registry::master(void) const
{
	if(m_hpMaster.isValid())
		return m_hpMaster;
	else
		return sp<Master>(NULL);
}

U32 Registry::prune(void)
{
	SAFEGUARD;

	U32 prunings=0;

	while(!m_loaderStack.empty())
	{
		LoadedLibrary &rLoadLib=*m_loaderStack.back();
		U32 adjusted=rLoadLib.m_spLibrary->adjustedReferences();
		U32 stray=rLoadLib.m_spLibrary->strayCount();
		U32 net=adjusted-stray;

		if(net!=rLoadLib.m_factoriesAccepted+1)
		{
			feLog("Registry::prune() \"%s\" has %d referencing objects\n"
					"  (incl. %d factories, %d hub strays, and 1 Registry)\n",
					rLoadLib.m_spLibrary->name().c_str(),
					adjusted,rLoadLib.m_factoriesAccepted,stray);

#if FE_COMPONENT_TRACK
			feLog("\nComponent %s\n",m_tracker.report().c_str());
#endif
		}

		FEASSERT(net>rLoadLib.m_factoriesAccepted);
		if(net-rLoadLib.m_factoriesAccepted != 1)
			return prunings;

		unmanage(m_loaderStack.size()-1);
		prunings++;
	}

	return prunings;
}

I32 Registry::findIndex(const String& library)
{
	SAFEGUARD;

	U32 size=m_loaderStack.size();
	U32 m;
	for(m=0;m<size;m++)
	{
		LoadedLibrary &rLoadLib= *(m_loaderStack[m]);
		if(rLoadLib.m_spLibrary->name()==library)
			return m;
	}
	return -1;
}

void Registry::dumpLibraries(void) const
{
	SAFEGUARD;

	U32 size=m_loaderStack.size();
	U32 m;
	for(m=0;m<size;m++)
	{
		LoadedLibrary &rLoadLib= *(m_loaderStack[m]);
		feLog("  \"%s\"  (%d %s)\n",
				rLoadLib.m_spLibrary->name().c_str(),
				rLoadLib.m_factoriesAccepted,
				rLoadLib.m_factoriesAccepted==1? "factory": "factories");
	}
}

void Registry::dump(void) const
{
	SAFEGUARD;

	printf("Registry::dump\n");
	std::map<String,FactoryLocation*>::const_iterator it=m_factoryMap.begin();
	while(it!=m_factoryMap.end())
	{
		printf("  factory %p \"%s\" in \"%s\"\n",
				it->second,it->first.c_str(),
				it->second->m_spLibrary->name().c_str());
		it++;
	}
}

Result Registry::unmanage(const String& library)
{
	SAFEGUARD;

	I32 index=findIndex(library);
	if(index<0)
		return e_cannotFind;

	unmanage(index);
	return e_ok;
}

void Registry::unmanage(U32 index)
{
	SAFEGUARD;

	LoadedLibrary* pLoadLib=m_loaderStack[index];
	String filename=pLoadLib->m_spLibrary->name();

#if FE_REGISTRY_MANAGE_DEBUG
	feLog("Registry::unmanage \"%s\"\n",filename.c_str());
#endif

	//* release the factories (is this too slow?)
	std::map<String,FactoryLocation*>::iterator it=m_factoryMap.begin();
	while(it!=m_factoryMap.end())
	{
		if(it->second->m_spLibrary->name()==filename)
		{
			FactoryLocation* pFactoryLocation=it->second;

#if FE_REGISTRY_FACTORY_DEBUG
			feLog("Registry::unmanage releasing factory %p \"%s\"\n",
					pFactoryLocation,it->first.c_str());
#endif

			delete pFactoryLocation;
			std::map<String,FactoryLocation*>::iterator it2=it++;
			m_factoryMap.erase(it2);
		}
		else
		{
			it++;
		}
	}

	if(pLoadLib->m_spLibrary->adjustedReferences()-
			pLoadLib->m_spLibrary->strayCount() != 1)
	{
		feLog("Registry::unmanage abandoning \"%s\""
				" still in use (%d extra ref)\n",
				pLoadLib->m_spLibrary->name().c_str(),
				pLoadLib->m_spLibrary->adjustedReferences()-1);
	}

	pLoadLib->m_spLibrary->releaseSingletons();
	pLoadLib->m_spLibrary->setRegistry(sp<Registry>(NULL));
	delete pLoadLib;

	U32 m;
	for(m=index;m<m_loaderStack.size()-1;m++)
		m_loaderStack[m]=m_loaderStack[m+1];
	m_loaderStack.pop_back();
}

void Registry::clear(void)
{
	SAFEGUARD;

	prune();

#if FALSE
	//* to speed things up, wipe out all lookups to avoid selective cleanup
	std::map<String,FactoryLocation*>::iterator it=m_factoryMap.begin();
	while(it!=m_factoryMap.end())
	{
		delete it->second;
		std::map<String,FactoryLocation*>::iterator it2=it++;
		m_factoryMap.erase(it2);
	}
#endif

	while(!m_loaderStack.empty())
		unmanage(m_loaderStack.size()-1);
}

Result Registry::registerLibrary(sp<Library> spLibrary)
{
	for(unsigned int i = 0; i < m_loaderStack.size(); i++)
	{
		if(m_loaderStack[i]->m_spLibrary->name() == spLibrary->name())
		{
			return e_alreadyAvailable;
		}
	}

	spLibrary->setRegistry(sp<Registry>(this));
	if(!m_loaderStack.empty())
		spLibrary->setChainLibrary(m_loaderStack.back()->m_spLibrary);

	LoadedLibrary *pLoadedLibrary=new LoadedLibrary(spLibrary);
	m_loaderStack.push_back(pLoadedLibrary);

	std::vector<String> manifest;
	spLibrary->populateManifest(manifest);
	U32 m;
	U32 size=manifest.size();
	for(m=0;m<size;m++)
	{
		const String &name=manifest[m];

#if FE_REGISTRY_FACTORY_DEBUG
		feLog("  registering factory \"%s\"\n",name.c_str());
#endif

		std::map<String,FactoryLocation*>::iterator it=m_factoryMap.find(name);
		if(it!=m_factoryMap.end())
		{
			FactoryLocation *pLocation=it->second;
			printf("WARNING redundant factory \"%s\"\n"
					"  previously registered by \"%s\"\n",
					name.c_str(),pLocation->m_spLibrary->name().c_str());
		}
		else
		{
			FactoryLocation* pFactoryLocation=new FactoryLocation(spLibrary,m);

			m_factoryMap[name]=pFactoryLocation;
			pLoadedLibrary->m_factoriesAccepted++;

#if FE_REGISTRY_FACTORY_DEBUG
			feLog("  %p\n",pFactoryLocation);
#endif
		}
	}

	return e_ok;
}

Result Registry::manage(const String &filename, bool adaptname,
		bool manageDependencies)
{
	SAFEGUARD;

	if(findIndex(filename)>=0)
		return e_alreadyAvailable;

	DL_Loader *pLoader=new DL_Loader();
	FEASSERT(pLoader);

#if FE_REGISTRY_MANAGE_DEBUG
	feLog("Registry::manage(\"%s\")\n",filename.c_str());
#endif

	BWORD success=pLoader->openDL(filename,adaptname);
	if(!success)
	{
		delete pLoader;
		return e_readFailed;
	}

	//* Load all dependencies first
	//* TODO prevent cycles (add current to an exclusion list)
	void (*fpListDependencies)(List<String*>&);
	fpListDependencies=(void (*)(List<String*>&))
			pLoader->findDLFunction("ListDependencies");

	if(manageDependencies && fpListDependencies)
	{
		//* NOTE Win32 sidestep: need to replace new'ed strings
		List<String*> preloadList;
		preloadList.setAutoDestruct(TRUE);
		fpListDependencies(preloadList);

#if TRUE
		List<String*>::Iterator iterator(preloadList);
		iterator.toHead();
		while(*iterator)
		{
			delete iterator.replace(new String((*iterator)->c_str()));
			iterator++;
		}
#endif

		BWORD unloaded=FALSE;
		String* pDependency;
		iterator.toHead();
		while((pDependency= *iterator++))
		{
#if FE_REGISTRY_MANAGE_DEBUG
			feLog("  \"%s\" requires dependency \"%s\"",filename.c_str(),
					pDependency->c_str());
#endif
			if(findIndex(*pDependency)<0)
			{
#if FE_REGISTRY_MANAGE_DEBUG
				feLog("  not loaded \n");
#endif
				if(!unloaded)
				{
					//* unload desired lib while we load dependencies
					success=pLoader->closeDL();
					FEASSERT(success);
					unloaded=TRUE;
				}
				manage(*pDependency);
			}
#if FE_REGISTRY_MANAGE_DEBUG
			else
			{
				feLog("  already loaded \n");
			}
#endif
		}

		if(unloaded)
		{
			//* reload desired lib
			success=pLoader->openDL(filename,adaptname);
			FEASSERT(success);
		}
	}

#if FE_REGISTRY_MANAGE_DEBUG
	feLog("Registry::manage \"%s\" loaded\n",filename.c_str());
#endif

	sp<TypeMaster> spTypeMaster;
	if(m_hpMaster.isValid())
		spTypeMaster=m_hpMaster->typeMaster();
	Library* (*fpCreateLibrary)(sp<Master>);
	fpCreateLibrary=(Library *(*)(sp<Master>))
			pLoader->findDLFunction("CreateLibrary");
	if(!fpCreateLibrary)
	{
		return e_invalidFile;
	}
	sp<Library> spLibrary(fpCreateLibrary(master()));
	spLibrary->setLoader(pLoader);

	for(unsigned int i = 0; i < m_loaderStack.size(); i++)
	{
		if(m_loaderStack[i]->m_spLibrary->name() == spLibrary->name())
		{
			return e_alreadyAvailable;
		}
	}

	if(!m_loaderStack.empty())
		spLibrary->setChainLibrary(m_loaderStack.back()->m_spLibrary);

	return manage(spLibrary);
}

Result Registry::manage(sp<Library>& a_rspLibrary)
{
	a_rspLibrary->setRegistry(sp<Registry>(this));

	LoadedLibrary *pLoadedLibrary=new LoadedLibrary(a_rspLibrary);
	m_loaderStack.push_back(pLoadedLibrary);

	std::vector<String> manifest;
	a_rspLibrary->populateManifest(manifest);
	U32 m;
	U32 size=manifest.size();
	for(m=0;m<size;m++)
	{
		const String &name=manifest[m];

#if FE_REGISTRY_FACTORY_DEBUG
		feLog("  registering factory \"%s\"\n",name.c_str());
#endif

		std::map<String,FactoryLocation*>::iterator it=m_factoryMap.find(name);
		if(it!=m_factoryMap.end())
		{
			FactoryLocation *pLocation=it->second;
			feLog("WARNING redundant factory \"%s\"\n"
					"  previously registered by \"%s\"\n",
					name.c_str(),pLocation->m_spLibrary->name().c_str());
		}
		else
		{
			FactoryLocation* pFactoryLocation=
					new FactoryLocation(a_rspLibrary,m);

			m_factoryMap[name]=pFactoryLocation;
			pLoadedLibrary->m_factoriesAccepted++;

#if FE_REGISTRY_FACTORY_DEBUG
			feLog("  %p\n",pFactoryLocation);
#endif
		}
	}

	return e_ok;
}

I32 Registry::listAvailable(const String &pattern,
	std::vector<String>& available)
{
	I32 count=0;

#if FE_REGISTRY_MATCH_DEBUG
	feLog("Registry::listAvailable(\"%s\") searching\n",pattern.c_str());
#endif

	std::map<String,FactoryLocation*>::const_iterator it=m_factoryMap.begin();
	while(it!=m_factoryMap.end())
	{
		const String &factoryName= it->first;

#if FE_REGISTRY_MATCH_DEBUG
		feLog("  vs. \"%s\"\n",factoryName.c_str());
#endif

		if(factoryName.dotMatch(pattern))
		{
#if FE_REGISTRY_MATCH_DEBUG
			feLog("  MATCHED\n");
#endif

			available.push_back(factoryName);
			count++;
		}
		it++;
	}

	return count;
}

void Registry::prioritize(const String &a_implementation,I32 a_priority)
{
	std::map<String,FactoryLocation*>::iterator it=
			m_factoryMap.find(a_implementation);
	if(it==m_factoryMap.end())
	{
		feLog("Registry::prioritize \"%s\" not found\n",
				a_implementation.c_str());
		return;
	}
	m_factoryMap[a_implementation]->m_priority=a_priority;
}

sp<Component> Registry::create(const String &a_pattern,BWORD quiet) const
{
	sp<Component> spComponent;

#if FE_REGISTRY_MATCH_DEBUG
	feLog("Registry::create(\"%s\") searching\n",a_pattern.c_str());
	quiet=FALSE;
#endif

	// TODO fast check for exact match
	// TODO store each new dot-match as a potential future exact match
	FactoryLocation *pBest=NULL;
	std::map<String,FactoryLocation*>::const_iterator it=m_factoryMap.begin();
	while(it!=m_factoryMap.end())
	{
		const String &factoryName= it->first;

#if FE_REGISTRY_MATCH_DEBUG
		feLog("  vs. \"%s\"\n",factoryName.c_str());
#endif

		if(factoryName.dotMatch(a_pattern))
		{
			FactoryLocation *pLocation=it->second;

#if FE_REGISTRY_MATCH_DEBUG
			feLog("  MATCHED %d\n",pLocation->m_priority);
#endif
			if(!pBest || pLocation->m_priority>pBest->m_priority)
			{
				sp<Component> spCandidate=pLocation->m_spLibrary->create(
						pLocation->m_factoryIndex);
				if(spCandidate.isValid())
				{
					pBest=pLocation;
					spComponent=spCandidate;
				}
			}
		}
		it++;
	}
	if(!quiet && !spComponent.isValid())
	{
		feLog("Registry::create(\"%s\") found no usable match\n",
				a_pattern.c_str());
	}

	return spComponent;
}

sp<Counted> Registry::create(String implementation,String name) const
{
	sp<Component> spComponent=create(implementation);
	spComponent->setName(name);
	return spComponent;
}

} /* namespace */
